import logging
from functools import partial
from typing import List, Optional

import orjson
from beartype import beartype

import picsellia.utils as utils
from picsellia.bcolors import bcolors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.model_version import ModelVersion
from picsellia.sdk.tag import Tag, TagTarget
from picsellia.sdk.taggable import Taggable
from picsellia.types.enums import Framework, InferenceType
from picsellia.types.schemas import ModelSchema

logger = logging.getLogger("picsellia")


class Model(Dao, Taggable):
    def __init__(self, connexion: Connexion, data: dict):
        Dao.__init__(self, connexion, data)
        Taggable.__init__(self, TagTarget.MODEL)

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> InferenceType:
        return self._type

    @property
    def framework(self) -> Framework:
        return self._framework

    def __str__(self):
        return "{}Model '{}' with type {} and framework {} {} (id: {})".format(
            bcolors.BLUE,
            self.name,
            self.type.name,
            self.framework.name,
            bcolors.ENDC,
            self.id,
        )

    @exception_handler
    @beartype
    def get_resource_url_on_platform(self) -> str:
        """Get platform url of this resource.

        Examples:
            ```python
                print(foo_dataset.get_resource_url_on_platform())
                >>> https://app.picsellia.com/model/62cffb84-b92c-450c-bc37-8c4dd4d0f590
            ```

        Returns:
            Url on Platform for this resource
        """

        return "{}/model/{}".format(self.connexion.host, self.id)

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get("/sdk/model/{}".format(self.id)).json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict):
        object = ModelSchema(**data)
        self._name = object.name
        self._type = object.type
        self._framework = object.framework
        return object

    @exception_handler
    @beartype
    def update(
        self,
        name: str = None,
        framework: Framework = None,
        private: bool = None,
        description: str = None,
        type: InferenceType = None,
    ) -> None:
        """Update a model with a new name, framework, privacy, description or type

        Examples:
            ```python
                model.update(description="Very cool model")
            ```
        """
        assert (
            type is None
            or type == InferenceType.CLASSIFICATION
            or type == InferenceType.OBJECT_DETECTION
            or type == InferenceType.SEGMENTATION
        ), "Model can only be a Classification, Detection or Segmentation"
        payload = {
            "name": name,
            "type": type,
            "framework": framework,
            "private": private,
            "description": description,
        }
        filtered_payload = utils.filter_payload(payload)
        r = self.connexion.patch(
            "/sdk/model/{}".format(self.id), data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)
        logger.info("{} updated.".format(self))

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete model.

        Delete the model in Picsellia database

        Examples:
            ```python
                model.delete()
            ```
        """
        self.connexion.delete("/sdk/model/{}".format(self.id))
        logger.info("{} deleted.".format(self))

    @exception_handler
    @beartype
    def get_tags(self) -> List[Tag]:
        """Retrieve the tags of your model.

        Examples:
            ```python
                tags = my_model.get_tags()
                assert tags[0].name == "my-model-1"
            ```

        Returns:
            List of tags as Tag
        """
        r = self.sync()
        return list(map(partial(Tag, self.connexion), r["tags"]))

    @exception_handler
    @beartype
    def create_version(
        self,
        docker_image_name: Optional[str] = None,
        docker_flags: Optional[List[str]] = None,
        thumb_object_name: Optional[str] = None,
        notebook_link: Optional[str] = None,
        labels: Optional[dict] = None,
        base_parameters: Optional[dict] = None,
    ) -> ModelVersion:
        """Create a version of a model.

        The version number of this model will be defined by the platform. It is incremented automatically.

        Examples:
            ```python
                model_v0 = model.create_version()
            ```

        Returns:
            A (ModelVersion) object
        """
        payload = {
            "docker_image_name": docker_image_name,
            "docker_flags": docker_flags,
            "thumb_object_name": thumb_object_name,
            "notebook_link": notebook_link,
            "labels": labels,
            "base_parameters": base_parameters,
        }
        filtered_payload = utils.filter_payload(payload)
        r = self.connexion.post(
            "/sdk/model/{}/versions".format(self.id),
            data=orjson.dumps(filtered_payload),
        ).json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def get_version(self, version: int) -> ModelVersion:
        """Retrieve a version of a model

        Examples:
            ```python
                version = model.get_version(0)
            ```

        Returns:
            A (ModelVersion) object
        """
        params = {"version": version}
        r = self.connexion.get(
            "/sdk/model/{}/versions/find".format(self.id), params=params
        ).json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def list_versions(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order_by: Optional[List[str]] = None,
    ) -> List[ModelVersion]:
        """List versions of a model

        Examples:
            ```python
            versions = model.list_versions()
            ```

        Returns:
            A list of (ModelVersion) object of this model
        """
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        params = utils.filter_payload(params)
        r = self.connexion.get(
            "/sdk/model/{}/versions".format(self.id), params=params
        ).json()
        return list(map(partial(ModelVersion, self.connexion), r["items"]))
